/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy;

import java.util.concurrent.atomic.AtomicBoolean;

import org.t2framework.commons.Disposer;
import org.t2framework.commons.module.Module;
import org.t2framework.commons.module.Modules;
import org.t2framework.commons.module.ModulesBootstrap;
import org.t2framework.commons.util.Assertion;
import org.t2framework.lucy.exception.LucyStartupFailException;
import org.t2framework.lucy.impl.DefaultLucyConfiguration;
import org.t2framework.lucy.impl.LucyImpl;
import org.t2framework.lucy.spi.LucyConfiguration;

/**
 * 
 * <#if locale="en">
 * <p>
 * Bootstrap class for Lucy.
 * </p>
 * <#else>
 * <p>
 * {@code Lucy}の起動クラスです．Lucyの初期化、生成( {@link #init(LucyConfiguration, Object)}
 * )、破棄({@link #destroy()})を行う事が出来ます．
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 */
public final class LucyBootstrap {

	/**
	 * <#if locale="en">
	 * <p>
	 * {@link Lucy} instance.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	private static Lucy lucy;

	/**
	 * <#if locale="en">
	 * <p>
	 * Modules.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	private static Modules modules = new Modules();

	/**
	 * <#if locale="en">
	 * <p>
	 * initialized flag.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	protected static AtomicBoolean initialized = new AtomicBoolean(false);

	/**
	 * <#if locale="en">
	 * <p>
	 * Initialize {@link Lucy}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @return
	 */
	public static Lucy init() {
		return init((String) null);
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Initialize {@link Lucy} with component definition path like xml path.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param componentDefinitionPath
	 * @return
	 */
	public static Lucy init(final String componentDefinitionPath) {
		return init(new DefaultLucyConfiguration(), componentDefinitionPath);
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Initialize {@link Lucy} with {@link LucyConfiguration}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param config
	 * @return
	 */
	public static Lucy init(final LucyConfiguration config) {
		return init(config, null);
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Initialize {@link Lucy} with {@link LucyConfiguration} and outer
	 * component definition path. The configuration instance must not be null.
	 * </p>
	 * <#else>
	 * <p>
	 * {@code Lucy}を初期化します.初期化時に{@code LucyConfiguration} とコンポーネントの定義を読み込み、
	 * {@code Lucy}を生成します.
	 * </p>
	 * </#if>
	 * 
	 * @param config
	 * @param componentDefinition
	 * @return
	 */
	public static synchronized Lucy init(final LucyConfiguration config,
			final Object componentDefinition) {
		Assertion.notNull(config);
		if (initialized.get()) {
			return lucy;
		}
		if (lucy == null) {
			modules.init();
			ModulesBootstrap.setModules(modules);
			lucy = new LucyImpl(config);
		}
		lucy.init();
		if (componentDefinition != null) {
			lucy.load(componentDefinition);
		}
		try {
			return lucy;
		} finally {
			initialized.set(true);
		}
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Load component definition.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param componentDefinition
	 */
	public static void load(final Object componentDefinition) {
		if (lucy == null) {
			throw new LucyStartupFailException();
		}
		lucy.load(componentDefinition);
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Get instantiated {@link Lucy}.If there is no {@link Lucy} instance, throw
	 * {@link LucyStartupFailException}.
	 * </p>
	 * <#else>
	 * <p>
	 * {@code Lucy}インスタンスを返します.インスタンスが生成されていない場合、{@code
	 * LucyStartupFailException}が発生します.
	 * </p>
	 * </#if>
	 * 
	 * @return
	 */
	public static Lucy getLucy() {
		if (lucy == null) {
			throw new LucyStartupFailException();
		}
		return lucy;
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Destroy instantiated {@link Lucy} instance, and
	 * {@link Disposer#dispose()} resources.
	 * </p>
	 * <#else>
	 * <p>
	 * {@code Lucy}を破棄します.Lucy自身の破棄後に{@link Disposer#dispose()} を呼び出します.
	 * </p>
	 * </#if>
	 * 
	 */
	public static synchronized void destroy() {
		if (lucy != null) {
			modules.dispose();
			lucy.destroy();
			lucy = null;
		}
		Disposer.dispose();
		initialized.compareAndSet(true, false);
	}

	public static <M extends Module> M getModule(Class<M> moduleClass) {
		return (M) getModules().getModule(moduleClass);
	}

	public static Modules getModules() {
		return modules;
	}

	public static <S> S getService(final Class<? super S> serviceClass) {
		return (S) getModules().getService(serviceClass);
	}

	public static void addModule(Module... modules) {
		getModules().addModule(modules);
	}

	public synchronized static void setModules(Modules modules) {
		LucyBootstrap.modules = modules;
		ModulesBootstrap.setModules(modules);
	}
}
