/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.aop.impl;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.t2framework.commons.aop.spi.Aspect;
import org.t2framework.commons.aop.spi.Interceptor;
import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.BeanDescUtil;
import org.t2framework.commons.meta.Config;
import org.t2framework.commons.meta.ConfigContainer;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.util.Assertion;
import org.t2framework.commons.util.CollectionsUtil;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.aop.AspectCreator;
import org.t2framework.lucy.aop.config.AspectConfig;
import org.t2framework.lucy.aop.exception.InterceptorNotFoundRuntimeException;

public class AspectCreationHelper {

	public static <T, C> AspectAssemblerHelperChain<T> build(
			final AspectCreator creator) {
		return new AspectAssemblerHelperChain<T>() {

			protected LinkedList<ChainInvoker<T>> invokers = new LinkedList<ChainInvoker<T>>();

			@Override
			public List<Aspect> doChain(final BeanDesc<T> beanDesc,
					final Lucy lucy) {
				LinkedList<Aspect> ret = new LinkedList<Aspect>();
				for (ChainInvoker<T> invoker : invokers) {
					List<Aspect> list = invoker.invoke(beanDesc, lucy);
					if (list != null && !list.isEmpty()) {
						ret.addAll(list);
					}
				}
				return ret;
			}

			protected AspectAssemblerHelperChain<T> setup() {
				invokers.add(new ChainInvoker<T>() {

					protected final MethodAspectAssemblerHelper helper = new MethodAspectAssemblerHelper(
							creator);

					@Override
					public List<Aspect> invoke(BeanDesc<T> beanDesc, Lucy lucy) {
						return helper
								.assemble(
										BeanDescUtil
												.getMethodConfigContainerList(
														beanDesc,
														org.t2framework.commons.annotation.core.Aspect.class),
										lucy);
					}

				});
				invokers.add(new ChainInvoker<T>() {

					protected final AspectConfigAssemblerHelper helper = new AspectConfigAssemblerHelper();

					@Override
					public List<Aspect> invoke(BeanDesc<T> beanDesc, Lucy lucy) {
						List<ConfigContainer> list = new ArrayList<ConfigContainer>();
						list.add(beanDesc.getClassDesc());
						return helper.assemble(list, lucy);
					}

				});
				invokers.add(new ChainInvoker<T>() {

					protected final DefaultAspectAssemblerHelper helper = new DefaultAspectAssemblerHelper(
							creator);

					@Override
					public List<Aspect> invoke(BeanDesc<T> beanDesc, Lucy lucy) {
						List<ConfigContainer> list = new ArrayList<ConfigContainer>();
						list.add(beanDesc.getClassDesc());
						return helper.assemble(list, lucy);
					}

				});
				return this;
			}
		}.setup();
	}

	public static interface ChainInvoker<T> {
		List<Aspect> invoke(BeanDesc<T> beanDesc, Lucy lucy);
	}

	public static interface AspectAssemblerHelperChain<T> {

		List<Aspect> doChain(final BeanDesc<T> beanDesc, final Lucy lucy);
	}

	public static interface AspectAssembleHelper {

		List<Aspect> assemble(Iterable<ConfigContainer> list, Lucy lucy);
	}

	public static class DefaultAspectAssemblerHelper implements
			AspectAssembleHelper {

		protected AspectCreator creator;

		public DefaultAspectAssemblerHelper(AspectCreator creator) {
			this.creator = Assertion.notNull(creator);
		}

		@Override
		public List<Aspect> assemble(Iterable<ConfigContainer> iterable,
				Lucy lucy) {
			if (iterable == null) {
				return null;
			}
			List<Aspect> ret = CollectionsUtil.newArrayList();
			for (ConfigContainer container : iterable) {
				Config config = container
						.findConfig(org.t2framework.commons.annotation.core.Aspect.class);
				if (config == null) {
					continue;
				}
				org.t2framework.commons.annotation.core.Aspect aspect = getAspect(config);
				final String regex = getRegex(aspect);
				List<Interceptor> interceptors = getInterceptors(aspect, lucy);
				ret.add(createAspect(container, config, regex, interceptors));
			}
			return ret;
		}

		protected String getRegex(
				org.t2framework.commons.annotation.core.Aspect aspect) {
			return aspect.pointcut();
		}

		protected List<Interceptor> getInterceptors(
				org.t2framework.commons.annotation.core.Aspect aspect, Lucy lucy) {
			List<Interceptor> interceptors = CollectionsUtil.newArrayList();
			Class<? extends Interceptor>[] classes = aspect.interceptBy();
			for (int j = 0; j < classes.length; j++) {
				final Class<? extends Interceptor> clazz = classes[j];
				final Interceptor interceptor = lucy.get(clazz);
				if (interceptor == null) {
					throw new InterceptorNotFoundRuntimeException(clazz);
				}
				interceptors.add(interceptor);
			}
			return interceptors;
		}

		protected Aspect createAspect(ConfigContainer container, Config conf,
				final String regex, final List<Interceptor> interceptors) {
			return creator.createAspect(regex, interceptors);
		}

		public AspectCreator getCreator() {
			return creator;
		}

	}

	public static class MethodAspectAssemblerHelper extends
			DefaultAspectAssemblerHelper {

		public MethodAspectAssemblerHelper(AspectCreator creator) {
			super(creator);
		}

		@Override
		protected Aspect createAspect(ConfigContainer container, Config conf,
				String regex, List<Interceptor> interceptors) {
			if (MethodDesc.class.isInstance(container)) {
				MethodDesc md = MethodDesc.class.cast(container);
				return getCreator().createAspect(md.getMethodName(),
						interceptors);
			} else {
				throw new IllegalStateException("aspect should be on Method");
			}
		}

	}

	public static class AspectConfigAssemblerHelper implements
			AspectAssembleHelper {

		@Override
		public List<Aspect> assemble(
				Iterable<ConfigContainer> configContainerList, Lucy lucy) {
			if (configContainerList == null) {
				return null;
			}
			List<Aspect> retList = CollectionsUtil.newArrayList();
			for (ConfigContainer container : configContainerList) {
				Iterable<Config> aspectConfigs = container
						.findConfigs(AspectConfig.class);
				if (aspectConfigs == null) {
					continue;
				}
				for (Config config : aspectConfigs) {
					retList.add(((AspectConfig) config).createAspect());
				}
			}
			return retList;
		}

	}

	protected static org.t2framework.commons.annotation.core.Aspect getAspect(
			Config config) {
		if (config.hasAnnotation() == false) {
			return null;
		}
		return (org.t2framework.commons.annotation.core.Aspect) config
				.getAnnotation();
	}

}
