/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.aop.javassist;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;

import org.t2framework.commons.util.Assertion;
import org.t2framework.lucy.aop.AspectDesc;
import org.t2framework.lucy.aop.exception.CannotCompileRuntimeException;
import org.t2framework.lucy.aop.javassist.util.ClassPoolUtil;
import org.t2framework.lucy.aop.javassist.util.JavassistUtil;
import org.t2framework.lucy.aop.javassist.util.LucyClassPool;

/**
 * <#if locale="en">
 * <p>
 * 
 * </p>
 * <#else>
 * <p>
 * 
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 */
public class ClassEnhanceHelper {

	/**
	 * <#if locale="en">
	 * <p>
	 * Make {@link CtClass}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param aspectDesc
	 * @return
	 */
	public static CtClass makeCtClass(final AspectDesc aspectDesc) {
		Assertion.notNull(aspectDesc);
		final String enhancedClassName = aspectDesc.getEnhancedClassName();
		final Class<?> originalClass = aspectDesc.getOriginalClass();
		final ClassPool pool = ClassPoolUtil.getClassPool(originalClass);

		boolean isInterface = aspectDesc.isInterfaceTarget();
		Class<?> clazz = isInterface ? Object.class : originalClass;
		CtClass ctClass = JavassistUtil.makeCtClass(pool, enhancedClassName,
				clazz);
		if (isInterface) {
			CtClass originalCtClass = JavassistUtil.getCtClass(pool,
					originalClass);
			ctClass.setInterfaces(new CtClass[] { originalCtClass });
		} else {
			Class<?>[] interfaces = originalClass.getInterfaces();
			CtClass[] ctClassArray = JavassistUtil.getCtClassArray(pool,
					interfaces);
			ctClass.setInterfaces(ctClassArray);
		}
		setupConstructors(pool, ctClass, originalClass);
		return ctClass;
	}

	protected static void setupConstructors(ClassPool pool, CtClass ctClass,
			Class<?> originalClass) {
		Constructor<?>[] constructors = originalClass.getDeclaredConstructors();
		if (constructors.length == 0) {
			JavassistUtil.addDefaultConstructor(ctClass);
		} else {
			for (Constructor<?> constructor : constructors) {
				if (shouldCreateCtConstructor(originalClass, constructor)) {
					createCtConstructor(pool, ctClass, constructor);
				}
			}
		}
	}

	protected static void createCtConstructor(ClassPool pool, CtClass ctClass,
			Constructor<?> constructor) {
		CtClass[] paramTypes = JavassistUtil.getCtClassArray(pool, constructor
				.getParameterTypes());
		CtClass[] exceptionTypes = JavassistUtil.getCtClassArray(pool,
				constructor.getExceptionTypes());
		try {
			final CtConstructor ctConstructor = CtNewConstructor.make(
					paramTypes, exceptionTypes, ctClass);
			ctClass.addConstructor(ctConstructor);
		} catch (final CannotCompileException e) {
			throw new CannotCompileRuntimeException(e);
		}
	}

	protected static boolean shouldCreateCtConstructor(Class<?> originalClass,
			Constructor<?> constructor) {
		final int modifier = constructor.getModifiers();
		final String classname = originalClass.getName();
		final Package p = originalClass.getPackage();
		final boolean isPublicOrProtected = Modifier.isPublic(modifier)
				|| Modifier.isProtected(modifier)
				|| Modifier.isPrivate(modifier) == false;
		final boolean notJavaPackage = classname.indexOf("java") == 0;
		final boolean packageNotSealed = (p == null || p.isSealed() == false);
		return isPublicOrProtected && notJavaPackage && packageNotSealed;
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Add new {@link CtMethod}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param aspectDesc
	 * @param enhancedCtClass
	 * @param method
	 * @param methodName
	 * @param body
	 */
	public static void addCtNewMethod(final AspectDesc aspectDesc,
			final CtClass enhancedCtClass, final Method method,
			final String methodName, final String body) {
		LucyClassPool pool = ClassPoolUtil.getClassPool(aspectDesc
				.getOriginalClass());
		CtMethod ctMethod = LucyCtNewMethod.make(enhancedCtClass, pool, method,
				methodName, body);
		try {
			enhancedCtClass.addMethod(ctMethod);
		} catch (CannotCompileException e) {
			throw new CannotCompileRuntimeException(e);
		}
	}

	public static class LucyCtNewMethod extends CtNewMethod {

		public static CtMethod make(CtClass enhancedCtClass,
				LucyClassPool pool, Method method, String methodName,
				String body) {
			final CtClass returnType = JavassistUtil.getCtClass(pool, method
					.getReturnType());
			final CtClass[] parameters = JavassistUtil.getCtClassArray(pool,
					method.getParameterTypes());
			final CtClass[] exceptions = JavassistUtil.getCtClassArray(pool,
					method.getExceptionTypes());
			try {
				CtMethod cm = new CtMethod(returnType, methodName, parameters,
						enhancedCtClass);
				final int mod = method.getModifiers()
						& ~(Modifier.ABSTRACT | Modifier.NATIVE);
				cm.setModifiers(mod);
				cm.setExceptionTypes(exceptions);
				cm.setBody(body);
				return cm;
			} catch (Exception e) {
				throw new CannotCompileRuntimeException(e);
			}
		}

	}
}
