/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.aop.javassist;

import java.lang.reflect.AccessibleObject;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import javassist.CannotCompileException;
import javassist.CtClass;
import javassist.CtField;
import javassist.Modifier;

import org.t2framework.commons.aop.spi.Interceptor;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.util.Assertion;
import org.t2framework.commons.util.Reflections.ClassLoaderUtil;
import org.t2framework.lucy.aop.AspectDesc;
import org.t2framework.lucy.aop.InterceptorDesc;
import org.t2framework.lucy.aop.InterceptorDescStrategy;
import org.t2framework.lucy.aop.exception.CannotCompileRuntimeException;
import org.t2framework.lucy.aop.javassist.util.JavassistUtil;

/**
 * 
 * <#if locale="en">
 * <p>
 * Javassist based enhance class builder.
 * </p>
 * <#else>
 * <p>
 * 
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 * @param <T>
 */
public class JavassistEnhancedClassBuilderImpl<T> extends
		AbstractEnhancedClassBuilder<T> {

	public JavassistEnhancedClassBuilderImpl(InterceptorDescStrategy strategy) {
		super(strategy);
	}

	@Override
	protected CtClass makeCtClass(AspectDesc aspectDesc) {
		return ClassEnhanceHelper.makeCtClass(aspectDesc);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected Class<T> generateClass(final AspectDesc aspectDesc,
			CtClass enhancedCtClass) {
		Assertion.notNulls(aspectDesc, enhancedCtClass);
		final ClassLoader classLoader = ClassLoaderUtil
				.getClassLoader(aspectDesc.getOriginalClass());
		try {
			return (Class<T>) JavassistUtil.toClass(enhancedCtClass,
					classLoader);
		} finally {
			enhancedCtClass.detach();
			enhancedCtClass = null;
		}
	}

	@Override
	protected void createInterceptorsField(String interceptorsName,
			Class<?> originalClass, CtClass declaringCtClass,
			Map<AccessibleObject, List<Interceptor>> interceptorMap) {
		try {
			CtField ctField = JavassistUtil.makeCtField(interceptorsName,
					originalClass, interceptorMap, declaringCtClass);
			JavassistUtil.setCtFieldAsPublicStatic(ctField);
			declaringCtClass.addField(ctField);
		} catch (CannotCompileException e) {
			throw new CannotCompileRuntimeException(e);
		}
	}

	@Override
	protected void createInvokeSuperMethod(AspectDesc aspectDesc,
			CtClass enhancedCtClass,
			Collection<InterceptorDesc> interceptorDescList) {
		for (InterceptorDesc id : Assertion.notNull(interceptorDescList)) {
			final MethodDesc md = id.getMethodDesc();
			if (md == null || md.isAbstract()) {
				continue;
			}
			final String methodName = md.getMethodName();
			final String generatedName = generateMethodName(methodName);
			final String body = MethodTemplateFactory
					.getSuperMethodBody(methodName);
			ClassEnhanceHelper.addCtNewMethod(aspectDesc, enhancedCtClass, md
					.getMethod(), generatedName, body);
		}
	}

	protected String generateMethodName(String methodName) {
		return methodName + "$$";
	}

	@Override
	protected String generateInterceptorsFieldName(Class<?> clazz) {
		String name = "$$INTERCEPTORS$$";
		int i = 0;
		while (true) {
			try {
				clazz.getDeclaredField(name);
			} catch (NoSuchFieldException e) {
				try {
					clazz.getField(name);
				} catch (NoSuchFieldException ex) {
					return name;
				}
			}
			name = name + Integer.toString(i++);
		}
	}

	@Override
	protected void createTargetMethod(AspectDesc aspectDesc,
			CtClass enhancedCtClass,
			Map<AccessibleObject, InterceptorDesc> map, String interceptorsName) {
		for (InterceptorDesc id : Assertion.notNull(map).values()) {
			final MethodDesc md = id.getMethodDesc();
			if (md == null) {
				continue;
			}
			String methodName = md.getMethodName();
			String body = null;
			final boolean isAbstract = Modifier.isAbstract(md.getModifiers());
			if (isAbstract) {
				body = MethodTemplateFactory.getAbstractMethodBody(methodName,
						interceptorsName);
			} else {
				body = MethodTemplateFactory.getMethodBody(methodName,
						interceptorsName);
			}
			ClassEnhanceHelper.addCtNewMethod(aspectDesc, enhancedCtClass, md
					.getMethod(), methodName, body);
		}
	}

}
