/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.inject;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

import org.t2framework.commons.meta.Config;
import org.t2framework.commons.meta.ConfigContainer;
import org.t2framework.commons.meta.ExpressionAware;
import org.t2framework.commons.util.Assertion;
import org.t2framework.commons.util.AutoboxingUtil;
import org.t2framework.commons.util.ConverterUtil;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.annotation.core.Inject;
import org.t2framework.lucy.config.meta.InjectConfig;
import org.t2framework.lucy.el.ExpressionUtil;

/**
 * 
 * <#if locale="en">
 * <p>
 * Base class of {@link Injector}.This class gives basic information for
 * injection, which are {@link ConfigContainer} parameterized as C and
 * {@link InjectInfo}, to a subclass like {@link MethodInjector}.
 * </p>
 * <#else>
 * <p>
 * 汎用の抽象{@link Injector}クラスです．インジェクションに必要な情報を収集し、{@link InjectInfo} の形式にします．
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 * @param <C>
 */
public abstract class AbstractInjector<C extends ConfigContainer> implements
		Injector {

	/**
	 * <#if locale="en">
	 * <p>
	 * {@link ConfigContainer}.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	protected C container;

	/**
	 * <#if locale="en">
	 * <p>
	 * Injection information.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	protected InjectInfo injectInfo;

	/**
	 * <#if locale="en">
	 * <p>
	 * True if trying injection by annotation.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	protected boolean injectByAnnotation;

	/**
	 * <#if locale="en">
	 * <p>
	 * True if trying injection by configuration.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	protected boolean injectByConfiguration;

	public AbstractInjector(C container) {
		this.container = Assertion.notNull(container);
		setup(container);
	}

	protected void setup(C container) {
		this.injectByAnnotation = InjectorUtil.hasInjectAnnotation(container);
		this.injectByConfiguration = InjectorUtil
				.hasInjectConfiguration(container);
		setInjectInfo(new InjectInfo(container));
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Get the target {@link ConfigContainer}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @return
	 */
	protected C getConfigContainer() {
		return container;
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Get {@link InjectInfo}.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @return
	 */
	protected InjectInfo getInjectInfo() {
		return injectInfo;
	}

	protected void setInjectInfo(InjectInfo injectInfo) {
		this.injectInfo = injectInfo;
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Utility of converting types for subclass.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param obj
	 * @param componentType
	 * @return
	 */
	protected Object[] convertTypes(Object obj, Class<?> componentType) {
		final int length = Array.getLength(obj);
		Object[] values = (Object[]) Array.newInstance(componentType, length);
		for (int i = 0; i < length; i++) {
			Object convert = ConverterUtil.convert(Array.get(obj, i),
					componentType);
			Array.set(values, i, convert);
		}
		return values;
	}

	protected Object[] evaluateExpression(
			final ExpressionAware expressionAware, final Lucy lucy) {
		return ExpressionUtil.evaluateExpression(expressionAware, lucy);
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Prepare arguments for component injection.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param types
	 *            Array of Class<?> for looking up components from Lucy.This
	 *            must not be null.
	 * @param lucy
	 *            Lucy instance.Is must not be null.
	 * @param info
	 *            Injection optional information for getting components. It may
	 *            be null.
	 * @return
	 */
	@SuppressWarnings("unchecked")
	protected Object[] prepareArgsByTypes(final Class<?>[] types,
			final Lucy lucy, final InjectInfo info) {
		Assertion.notNull(types, "types");
		Assertion.notNull(lucy, "lucy");
		Object[] args = null;
		final int typesLength = types.length;
		final Object[] values = new Object[typesLength];
		for (int i = 0; i < typesLength; i++) {
			Class type = types[i];
			Object value = InjectorUtil.getInjectableComponent(type, info, i,
					new InjectorUtil.InjectHandler<Object>() {

						@Override
						public Object getFromIdentifier(Object identifier) {
							return lucy.get(identifier);
						}

						@Override
						public Object getFromType(Class type) {
							return (type.isArray() == false) ? lucy.get(type)
									: lucy.getAll(type.getComponentType());
						}
					});
			values[i] = AutoboxingUtil.getDefaultValueIfNull(types[i], value);
		}
		if (values != null && 0 < values.length) {
			args = values;
		}
		return args;
	}

	protected <T> boolean isProxyClass(T t) {
		return (t != null && Proxy.isProxyClass(t.getClass()));
	}

	protected <T> void invokeProxyInvocation(T t, Method method, Object[] args) {
		try {
			Proxy.getInvocationHandler(t).invoke(t, method, args);
		} catch (Throwable th) {
			// TODO error handling.
			throw new IllegalStateException(th);
		}
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Injection information.It has names and parameter types if they are given
	 * by annotation or xml configuration.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @author shot
	 * 
	 */
	protected static class InjectInfo {

		protected String[] names;

		protected Class<?>[] paramTypes;

		public InjectInfo(ConfigContainer container) {
			Config config = container.findConfig(InjectConfig.class);
			if (config == null) {
				config = container.findConfig(Inject.class);
			}
			setup(config);
		}

		/**
		 * <#if locale="en">
		 * <p>
		 * Get name if given by {@link ConfigContainer}.Null or blank treats as
		 * nothing given.
		 * 
		 * </p>
		 * <#else>
		 * <p>
		 * 
		 * </p>
		 * </#if>
		 * 
		 * @return
		 */
		public String[] getNames() {
			return this.names;
		}

		/**
		 * <#if locale="en">
		 * <p>
		 * Get parameter types if given by {@link ConfigContainer}.If parameter
		 * types is single Object.class parameter, it treats as nothing given.
		 * 
		 * </p>
		 * <#else>
		 * <p>
		 * 
		 * </p>
		 * </#if>
		 * 
		 * @return
		 */
		public Class<?>[] getParameterTypes() {
			return paramTypes;
		}

		/**
		 * <#if locale="en">
		 * <p>
		 * If there is name given.
		 * 
		 * </p>
		 * <#else>
		 * <p>
		 * 
		 * </p>
		 * </#if>
		 * 
		 * @return
		 */
		public boolean hasNames() {
			if (names == null) {
				return false;
			}
			return names.length != Inject.DEFAULT_NAME_LENGTH
					|| Inject.DEFAUL_NAME.equals(names[0]) == false;
		}

		/**
		 * <#if locale="en">
		 * <p>
		 * If there is parameter types given.
		 * 
		 * </p>
		 * <#else>
		 * <p>
		 * 
		 * </p>
		 * </#if>
		 * 
		 * @return
		 */
		public boolean hasParameterTypes() {
			final int length = paramTypes.length;
			if (paramTypes == null || length == 0) {
				return false;
			}
			return (1 < length)
					|| (length == 1 && paramTypes[0] != Object.class);
		}

		private void setup(Config config) {
			if (config instanceof InjectConfig) {
				InjectConfig injectConfig = (InjectConfig) config;
				this.paramTypes = injectConfig.getParameterTypes();
			} else {
				if (InjectorUtil.isInjectAnnotation(config) == false) {
					throw new IllegalStateException();
				}
				Annotation a = config.getAnnotation();
				if (a instanceof Inject) {
					Inject inject = (Inject) a;
					this.names = inject.name();
					this.paramTypes = inject.types();
				}
			}
		}
	}
}
