/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.inject;

import java.lang.reflect.Constructor;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.t2framework.commons.Constants;
import org.t2framework.commons.exception.ConstructorNotFoundRuntimeException;
import org.t2framework.commons.meta.BeanDesc;
import org.t2framework.commons.meta.ConfigContainer;
import org.t2framework.commons.meta.ConstructorDesc;
import org.t2framework.commons.util.Assertion;
import org.t2framework.commons.util.CollectionsUtil;
import org.t2framework.commons.util.Reflections.ConstructorUtil;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.annotation.core.Inject;
import org.t2framework.lucy.config.meta.InjectConfig;
import org.t2framework.lucy.config.meta.impl.ConstructorInjectConfig;
import org.t2framework.lucy.exception.ConstrucorInjectionFailureException;
import org.t2framework.lucy.exception.CyclicReferenceRuntimeException;

/**
 * 
 * <#if locale="en">
 * <p>
 * {@code ConstructorInjector} injects components using {@code Constructor} .
 * </p>
 * <#else>
 * <p>
 * コンストラクタインジェクションを実現するインジェクタクラスです．
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 */
public class ConstructorInjector extends AbstractInjector<BeanDesc<?>> {

	protected List<Constructor<?>> injectableConstructors = CollectionsUtil
			.newArrayList();

	public ConstructorInjector(BeanDesc<?> container) {
		super(Assertion.notNull(container));
		findInjectableConstructors(getConfigContainer().getConstructorDesc());
	}

	/**
	 * 
	 * <#if locale="en">
	 * <p>
	 * Injects components using {@code Constructor} .
	 * </p>
	 * <#else>
	 * <p>
	 * コンストラクタを使って、インジェクションを実現します．インジェクト対象のコンストラクタは引数の多い順序で探していき、
	 * コンポーネント解決がすべて出来たものを採用してインスタンス化します．
	 * </p>
	 * </#if>
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <T> T inject(T t, Lucy lucy) throws CyclicReferenceRuntimeException,
			ConstrucorInjectionFailureException {
		assertInjectableConstructorsExists();
		Object[] args = Constants.EMPTY_ARRAY;
		final InjectInfo injectInfo = getInjectInfo();
		for (Constructor<?> constructor : injectableConstructors) {
			if (ConstructorUtil.isDefaultConstructor(constructor)) {
				args = Constants.EMPTY_ARRAY;
				break;
			}
			final Class<?>[] paramTypes = getParameterTypes(constructor,
					injectInfo);
			args = prepareArgs(paramTypes, getConfigContainer()
					.getConstructorDesc(), lucy, injectInfo);
			if (args != null && Assertion.hasNotNull(args)) {
				break;
			}
			args = null;
		}
		if (args == null) {
			throw new ConstrucorInjectionFailureException(
					getConfigContainerName());
		}
		return (T) getConfigContainer().newInstance(args);
	}

	protected Object[] prepareArgs(Class<?>[] types,
			ConstructorDesc<?> constructorDesc, Lucy lucy, InjectInfo info) {
		Object[] args = null;
		if (constructorDesc.hasExpression()) {
			args = evaluateExpression(constructorDesc, lucy);
		} else {
			args = super.prepareArgsByTypes(types, lucy, info);
		}
		return args;
	}

	protected Class<?>[] getParameterTypes(Constructor<?> constructor,
			InjectInfo info) {
		return (info != null && info.hasParameterTypes()) ? info
				.getParameterTypes() : constructor.getParameterTypes();
	}

	protected void assertInjectableConstructorsExists() {
		if (injectableConstructors.size() < 1) {
			throw new ConstructorNotFoundRuntimeException(
					getConfigContainerName());
		}
	}

	protected String getConfigContainerName() {
		return getConfigContainer().getConcreteClass().getName();
	}

	protected boolean findInjectableConstructors(
			ConstructorDesc<?> constructorDesc) {
		if (constructorDesc.hasOnlyDefaultConstructor()) {
			addDefaultConstructorAsLast(constructorDesc);
			return true;
		}
		if (injectByAnnotation == false && injectByConfiguration == false) {
			return false;
		}
		ConstructorInjectConfig config = getConstructorInjectConfig(constructorDesc);
		for (Constructor<?> constructor : constructorDesc.getConstructors()) {
			// a synthetic constructor or a constructor which has var args is
			// not supported.
			if (constructor.isSynthetic() || constructor.isVarArgs()) {
				continue;
			}
			processConstructor(constructor, injectByAnnotation,
					injectByConfiguration, config);
		}
		if (injectByConfiguration == false
				&& constructorDesc.hasDefaultConstructor()) {
			addDefaultConstructorAsLast(constructorDesc);
		}
		Collections.sort(injectableConstructors, CONSTRUCTOR_COMPARATOR);
		return true;
	}

	protected void processConstructor(Constructor<?> constructor,
			boolean injectByAnnotation, boolean injectByConfiguration,
			ConstructorInjectConfig config) {
		if (config != null && config.isAutoinjected() == false) {
			final Class<?>[] paramTypes = constructor.getParameterTypes();
			if (ConstructorUtil.isDefaultConstructor(constructor)) {
				return;
			}
			if (isInjectAnnotationPresent(constructor) == false
					&& isInjectConfigurationArgsEquals(config, paramTypes) == false) {
				return;
			}
		}
		injectableConstructors.add(constructor);
	}

	protected boolean isInjectAnnotationPresent(Constructor<?> constructor) {
		return this.injectByAnnotation
				&& constructor.isAnnotationPresent(Inject.class);
	}

	protected boolean isInjectConfigurationArgsEquals(
			ConstructorInjectConfig config, Class<?>[] paramTypes) {
		return this.injectByConfiguration
				&& config.getArgumentSize() == paramTypes.length;
	}

	protected void addDefaultConstructorAsLast(ConstructorDesc<?> container) {
		Constructor<?> defaultConstructor = container
				.getConstructor(Constants.EMPTY_CLASS_ARRAY);
		if (defaultConstructor != null) {
			injectableConstructors.add(defaultConstructor);
		}
	}

	protected ConstructorInjectConfig getConstructorInjectConfig(
			ConfigContainer constructorDesc) {
		return (ConstructorInjectConfig) constructorDesc
				.findConfig(InjectConfig.class);
	}

	@Override
	protected void setup(BeanDesc<?> container) {
		ConstructorDesc<?> constructorDesc = getConfigContainer()
				.getConstructorDesc();
		this.injectByAnnotation = InjectorUtil
				.hasInjectAnnotation(constructorDesc);
		this.injectByConfiguration = InjectorUtil
				.hasInjectConfiguration(constructorDesc);
		if (injectByAnnotation || injectByConfiguration) {
			setInjectInfo(new InjectInfo(constructorDesc));
		}
	}

	private static final Comparator<Constructor<?>> CONSTRUCTOR_COMPARATOR = new Comparator<Constructor<?>>() {

		public static final int SORT = 1;

		public static final int NOCHANGE = 0;

		public static final int NOCHANGE2 = 0;

		@Override
		public int compare(Constructor<?> o1, Constructor<?> o2) {
			if (o1 == null || o2 == null) {
				return 0;
			}
			final int len1 = o1.getParameterTypes().length;
			final int len2 = o2.getParameterTypes().length;
			if (len1 < len2) {
				return SORT;
			} else if (len1 == len2) {
				return NOCHANGE;
			} else {
				return NOCHANGE2;
			}
		}

	};

}
