/*
 * Copyright 2008-2009 the T2 Project ant the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.t2framework.lucy.inject;

import org.t2framework.commons.exception.IllegalAccessRuntimeException;
import org.t2framework.commons.exception.InvocationTargetRuntimeException;
import org.t2framework.commons.meta.MethodDesc;
import org.t2framework.commons.util.Logger;
import org.t2framework.commons.util.PrintableUtil;
import org.t2framework.lucy.Lucy;
import org.t2framework.lucy.exception.CyclicReferenceRuntimeException;

/**
 * 
 * <#if locale="en">
 * <p>
 * {@link MethodInjector} is a concrete {@link Injector} class for method
 * injection.
 * 
 * <pre>
 * MethodInjection is enables to inject these type of injection:
 *  - simply single argument such as void inject(Emp emp);
 *  - single array arguments such as void inject(Emp[] emp);
 *  - multiple arguments not including array ex. void inject(Emp emp, Dept dept);
 * </pre>
 * 
 * 
 * </p>
 * <#else>
 * <p>
 * メソッドインジェクションを実現します．
 * 
 * 
 * メソッドインジェクションでは下記のようなインジェクトが可能です:
 * 
 * <pre>
 *  - 単一引数のインジェクト（例：void inject(Emp emp);）
 *  - 配列を使った同一の型を持つ複数コンポーネントのインジェクト（例：void inject(Emp[] emp);）
 *  - 配列を含まない複数の引数の同時インジェクト（例：void inject(Emp emp, Dept dept);）
 * </pre>
 * 
 * </p>
 * </#if>
 * 
 * @author shot
 * 
 */
public class MethodInjector extends AbstractInjector<MethodDesc> {

	protected static Logger logger = Logger.getLogger(MethodInjector.class);

	protected final String targetClassName;

	protected final String targetMethodName;

	public MethodInjector(MethodDesc methodDesc) {
		super(methodDesc);
		this.targetClassName = methodDesc.getDeclaringClass().getName();
		this.targetMethodName = methodDesc.getMethodName();
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Inject to the instance using with {@link MethodDesc}.
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 */
	@Override
	public <T> T inject(T t, Lucy lucy) throws CyclicReferenceRuntimeException,
			InvocationTargetRuntimeException, IllegalAccessRuntimeException {
		final MethodDesc methodDesc = getConfigContainer();
		final InjectInfo info = getInjectInfo();
		final Class<?>[] types = getParameterTypes(methodDesc, info);
		final Object[] args = prepareArgs(types, methodDesc, lucy, info);
		injectMethodDesc(methodDesc, t, args, types);
		return t;
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Prepare method invocation arguments by expression if {@link MethodDesc}
	 * has, or by types and {@link Lucy} container.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param types
	 * @param methodDesc
	 * @param lucy
	 * @return
	 */
	protected Object[] prepareArgs(Class<?>[] types, MethodDesc methodDesc,
			Lucy lucy, InjectInfo info) {
		Object[] args = null;
		if (methodDesc.hasExpression()) {
			args = evaluateExpression(methodDesc, lucy);
		} else {
			args = super.prepareArgsByTypes(types, lucy, info);
		}
		return args;
	}

	/**
	 * <#if locale="en">
	 * <p>
	 * Method injection execution.
	 * 
	 * </p>
	 * <#else>
	 * <p>
	 * 
	 * </p>
	 * </#if>
	 * 
	 * @param <T>
	 * @param methodDesc
	 * @param t
	 * @param args
	 * @param types
	 */
	protected <T> void injectMethodDesc(MethodDesc methodDesc, T t,
			Object[] args, Class<?>[] types) {
		if (isProxyClass(t)) {
			invokeProxyInvocation(t, methodDesc.getMethod(), args);
		} else {
			if (args == null) {
				logForNullInject(types);
				return;
			}
			if (1 < args.length || (args.length == 1 && args[0] != null)) {
				if (methodDesc.isVarArgs()
						&& this.injectInfo.hasParameterTypes()) {
					// TODO convert
					methodDesc.invoke(t, new Object[] { args });
				} else {
					methodDesc.invoke(t, args);
				}
			} else {
				logForNullInject(types);
			}
		}
	}

	protected void logForNullInject(Class<?>[] types) {
		logger.log("DLucyCore0001", new Object[] { targetClassName,
				targetMethodName, PrintableUtil.toPrintableString(types) });
	}

	protected Class<?>[] getParameterTypes(MethodDesc methodDesc,
			InjectInfo info) {
		return info.hasParameterTypes() ? info.getParameterTypes() : methodDesc
				.getParameterTypes();
	}

}
